@@ -6,6 +6,7 @@ import inspect
 import json
 import math
 import os
+import pickle
 
 if os.name == 'nt':
     import winreg
@@ -138,11 +139,26 @@ class WorkflowEngine:
         if as_dictionary:
             if not isinstance(self.input_parameter, dict):
                 if isinstance(self.input_parameter, str):
-                    self.input_parameter = self.input_parameter.replace("true", "True").replace("false", "False")
+                    self.input_parameter = self.input_parameter.replace("true", "True").replace("false", "False").replace("yes", "Yes").replace("no", "No")
                 self.input_parameter = eval(self.input_parameter)
         self.print_log(f"Got input parameter {str(self.input_parameter)}", "Processing Input")
         return self.input_parameter
 
+    def convert_binary_flow_to_default_file_format(self, flowpath: str) -> str:
+        """
+        Converts a flow to the default file format (base64 encoded json string) and saves it with the same name.
+        :param flowpath: The path to the binary flow that needs to be converted
+        """
+        # read the binary flow
+        flow = self.open(flowpath)
+        # base64 encode the content
+        string_bytes = json.dumps(flow).encode("ascii")
+        base64_bytes = base64.b64encode(string_bytes)
+        base64_string = base64_bytes.decode("ascii")
+        # write it back to the file
+        with open(flowpath, "w") as file:
+            file.write(base64_string)
+
     def open(self, filepath: str, as_xml: bool = False) -> any:
         """
         Open a DrawIO document
@@ -157,51 +173,65 @@ class WorkflowEngine:
         if not filepath.__contains__(".vsdx"):
             if filepath.__contains__(".flw"):
                 self.flowname = filepath.split("\\")[-1].replace(".flw", "")
-                with open(filepath, "rb") as binary_file:
-                    binary_file.seek(24)
-                    # Read the whole file at once
-                    content = binary_file.read()
-                str_content = content.decode("ascii", errors='ignore')
-                decoded = None
                 try:
-                    decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
-                    if decoded.__contains__("}]@"):
-                        decoded = decoded[0:decoded.index("}]@")+2]
-                    if decoded is not None:
-                        if decoded.__contains__("}]"):
-                            decoded = decoded[0:decoded.rfind("}]") + 2]
-                except (ValueError, Exception):
-                    str_content = str_content[:-1]
-                    str_content = "".join([x for x in str_content if x != '' and x != '']).strip().strip('\x00').strip('\x01').strip('\x0b')
-                if decoded is None:
+                    with open(filepath, "r") as f:
+                        content = f.read()
+                    decoded = base64.b64decode(content).decode("ascii", errors='ignore')
+                    retn = json.loads(decoded)
+                    return retn
+                except UnicodeDecodeError as e:
+                    # Found non-text data in the file
+                    print(e)
+                    try:
+                        with open(filepath, "rb") as binary_file:
+                            binary_file.seek(24)
+                            # Read the whole file at once
+                            content = binary_file.read()
+                        str_content = content.decode("ascii", errors='ignore')
+                        decoded = None
+                    except Exception as e:
+                        print(e)
+                        pass
                     try:
                         decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
-                    except (ValueError, Exception):
-                        remain = math.ceil((len(str_content)/4) - int(len(str_content)/4))
-                        str_content = str_content[0:len(str_content)-remain]
-                        idx = 0
-                        if str_content.__contains__("fV0") and not str_content.__contains__("In1d"):
-                            str_content = str_content.split("fV0")[0][1:] + "fV0==="
-                        if str_content.__contains__("In1d"):
-                            if str_content.__contains__("In1d") + 4 > idx:
-                                idx = str_content.index("In1d") + 4
-                        while idx > 0 and str_content.__contains__("In1d") and not str_content.endswith("In1d"):
-                            idx = str_content.index("In1d") + 4
-                            str_content = str_content[0:idx]
+                        if decoded.__contains__("}]@"):
+                            decoded = decoded[0:decoded.index("}]@") + 2]
+                        if decoded is not None:
+                            if decoded.__contains__("}]"):
+                                decoded = decoded[0:decoded.rfind("}]") + 2]
+                    except Exception as e:
+                        print(e)
+                        str_content = str_content[:-1]
+                        str_content = "".join([x for x in str_content if x != '' and x != '']).strip().strip('\x00').strip('\x01').strip('\x0b')
+                    if decoded is None:
                         try:
                             decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
                         except (ValueError, Exception):
-                            for i in range(0, 4):
-                                str_content = str_content[0:-1]
-                                try:
-                                    decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
-                                    break
-                                except (ValueError, Exception):
-                                    pass
-                        if decoded.__contains__("}]"):
-                            decoded = decoded[0:decoded.rfind("}]") + 2]
-                dict_list = json.loads(decoded)
-                return dict_list
+                            remain = math.ceil((len(str_content) / 4) - int(len(str_content) / 4))
+                            str_content = str_content[0:len(str_content) - remain]
+                            idx = 0
+                            if str_content.__contains__("fV0") and not str_content.__contains__("In1d"):
+                                str_content = str_content.split("fV0")[0][1:] + "fV0==="
+                            if str_content.__contains__("In1d"):
+                                if str_content.__contains__("In1d") + 4 > idx:
+                                    idx = str_content.index("In1d") + 4
+                            while idx > 0 and str_content.__contains__("In1d") and not str_content.endswith("In1d"):
+                                idx = str_content.index("In1d") + 4
+                                str_content = str_content[0:idx]
+                            try:
+                                decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
+                            except (ValueError, Exception):
+                                for i in range(0, 4):
+                                    str_content = str_content[0:-1]
+                                    try:
+                                        decoded = base64.b64decode(str_content).decode("ascii", errors='ignore')
+                                        break
+                                    except (ValueError, Exception):
+                                        pass
+                            if decoded.__contains__("}]"):
+                                decoded = decoded[0:decoded.rfind("}]") + 2]
+                    dict_list = json.loads(decoded)
+                    return dict_list
             else:
                 self.flowname = filepath.split("\\")[-1].replace(".xml", "")
                 xml_root = ElTree.fromstring(xml_file.read())
@@ -485,9 +515,9 @@ class WorkflowEngine:
                     else:
                         return_none = False
                 if isinstance(val, str):
-                    if val == "True":
+                    if val == "True" or val == "Yes":
                         val = True
-                    elif val == "False":
+                    elif val == "False" or val == "No":
                         val = False
                     else:
                         if val.replace(".", "").isnumeric():
@@ -644,6 +674,10 @@ class WorkflowEngine:
         if return_none:
             return None
         else:
+            # replace all reserved keys
+            for key in mapping:
+                if key in ["id ", "type ", "name ", "description "]:
+                    mapping[key.replace(" ", "")] = mapping.pop(key)
             return mapping
 
     @staticmethod
@@ -1238,14 +1272,14 @@ class WorkflowEngine:
                 try:
                     conn = \
                         [x for x in outgoing_connector if
-                         (str(x.value).lower() == "true" and x.source == current_step.id)][0]
+                         ((str(x.value).lower() == "true" or str(x.value).lower() == "yes") and x.source == current_step.id)][0]
                 except (ValueError, Exception):
                     raise Exception("Your Exclusive Gateway doesn't contain a 'True' or 'False' sequence arrow output.")
             else:
                 try:
                     conn = \
                         [x for x in outgoing_connector if
-                         (str(x.value).lower() == "false" and x.source == current_step.id)][0]
+                         ((str(x.value).lower() == "false" or str(x.value).lower() == "no") and x.source == current_step.id)][0]
                 except (ValueError, Exception):
                     raise Exception("Your Exclusive Gateway doesn't contain a 'True' or 'False' sequence arrow output.")
             try:
@@ -1711,3 +1745,156 @@ class Visio:
         for k, v in properties.items():
             setattr(retn, k, v)
         return retn
+
+
+class ChecklistEngine:
+    def __init__(self, flow_name="", full_path_save_as="", input_parameter=""):
+        """
+        This Engine will start the flow and will svae the state of the flow after each step by pickeling the flow object.
+        A Checklist can run multiple copies of the original flow. Therefore you must give each instance a separate name.
+        :param flow_name: The full path of the flow to run, including the file name and extension. This may be the original flow (.flw) file when you are starting a first run of a new instance, or the full path to the instance of the flow with the saved state.
+        :param input_parameter: The input parameter for the flow.
+        :param full_path_save_as: The full path to save the flow instance to. This is used to save the state of the flow after each step. You can save the flow with another name than the original flow to create separate instances of the flow. It is advised to set this parameter when you are starting a new instance of the flow.
+        """
+        self.flow_name = flow_name
+        self.save_as = full_path_save_as
+        # Check if file exists. If not, throw exception
+        if not os.path.exists(self.flow_name) and self.flow_name != "":
+            raise FileNotFoundError(f"The flow file does not exist!")
+        try:
+            self.engine = WorkflowEngine(input_parameter=input_parameter)
+            doc = self.engine.open(fr"{self.flow_name}")
+            self.engine.doc = doc
+            self.steps = self.engine.get_flow(doc)
+            self.flow_name = self.save_as
+        except Exception as e:
+            try:
+                self.load_flow_state(flow_name)
+            except Exception as e:
+                try:
+                    self.load_flow_state(full_path_save_as)
+                    self.flow_name = self.save_as
+                except Exception as e:
+                    pass
+
+    def run_next_step(self, ask_permission=False, msgbox=True):
+        """
+        Loop the steps and save after each step.
+        :param steps: The steps to run
+        :param ask_permission: If True, the user will be asked to click the OK button to continue to the next step.
+        :param msgbox: If this is True and ask_permission is True, a message box will be shown to the user to click OK to continue to the next step. Otherwise, the user will be asked to enter Yes (y) or No (n) in the console.
+        :return: True or False
+        """
+        outputPreviousStep = None
+        step = self.steps[self.engine.step_nr]
+        if hasattr(step, "shape_description"):
+            if getattr(step, "shape_description") == "End event.":
+                os.remove(self.flow_name)
+                self.engine.print_log(f"Flow finished, instance '{self.flow_name}' removed.")
+                print(f"Flow finished, instance '{self.flow_name}' removed.")
+                exit(0)
+        try:
+            if self.engine.db is None:
+                db_path = self.engine.get_db_path()
+                self.engine.db = SQL(db_path)
+            tmp = outputPreviousStep  # Needed for gateway
+            outputPreviousStep = self.engine.run_flow(step, True)
+            if outputPreviousStep is None:
+                outputPreviousStep = tmp  # When gateway
+            step = self.engine.get_next_step(step, self.steps, outputPreviousStep)
+            if step is None:
+                os.remove(self.flow_name)
+                self.engine.print_log(f"Flow finished, instance '{self.flow_name}' removed.")
+                print(f"Flow finished, instance '{self.flow_name}' removed.")
+                exit(0)
+            self.save_flow_state()
+            if ask_permission and not getattr(step, "IsStart") and not ("gateway" in str(getattr(step, "type"))):
+                if not self.ask_permission_for_next_step(msgbox=msgbox):
+                    exit(0)
+        except Exception as e:
+            pass
+        return True
+
+    def save_flow_state(self, full_path_to_save_to=""):
+        """
+        This function will save the state of the flow to a pickle file
+        :param flow: The flow to save
+        :param full_path_to_save_to: The full path to save the instance of the flow to.
+        """
+        if full_path_to_save_to != "":
+            self.save_as = full_path_to_save_to
+        if not self.save_as:
+            flw = self.flow_name.replace(".flw", "")
+        else:
+            flw = self.save_as
+        self.engine.db = None
+        with open(f"{flw}", "wb") as f:
+            pickle.dump(self.engine, f)
+
+    def load_flow_state(self, flow_path=None):
+        """
+        This function will load the state of the flow from a pickle file
+        :param flow_path: The full path of the flow to load
+        """
+        if flow_path:
+            flw = flow_path
+        else:
+            if not self.save_as:
+                flw = self.flow_name.replace(".flw", "")
+            else:
+                flw = self.save_as
+        if flw == "":
+            raise Exception("The path of the flow to load is empty!")
+        with open(f"{flw}", "rb") as f:
+            self.engine = pickle.load(f)
+        db_path = self.engine.get_db_path()
+        self.engine.db = SQL(db_path)
+        self.flow_name = flw
+        self.steps = self.engine.get_flow(self.engine.doc)
+
+    def resume_flow(self, flow_path, ask_permission=False, msgbox=True):
+        """
+        This function will resume the flow from the last saved state.
+        :param flow_path: The full path of the flow instance to resume
+        :param ask_permission: If True, the user will be asked to click the OK button to continue to the next step.
+        :param msgbox: If this is True and ask_permission is True, a message box will be shown to the user to click OK to continue to the next step. Otherwise, the user will be asked to enter Yes (y) or No (n) in the console.
+        """
+        self.load_flow_state(flow_path)
+        while True:
+            if not self.run_next_step(ask_permission=ask_permission, msgbox=msgbox):
+                exit(0)
+
+    def run_flow(self, ask_permission=False, msgbox=True):
+        """
+        This function will resume the flow from the last saved state.
+        :param ask_permission: If True, the user will be asked to click the OK button to continue to the next step.
+        :param msgbox: If this is True and ask_permission is True, a message box will be shown to the user to click OK to continue to the next step. Otherwise, the user will be asked to enter Yes (y) or No (n) in the console.
+        """
+        while True:
+            if not self.run_next_step(ask_permission=ask_permission, msgbox=msgbox):
+                exit(0)
+
+    def ask_permission_for_next_step(self, msgbox=True):
+        """
+        This function will ask the user for permission to run the next step. If the user does not have permission, the flow will be saved and the program will exit.
+        :param msgbox: If True, a messagebox will be shown to the user. If False, the user will be asked to enter Yes (y) or No (n) in the console.
+        """
+        instance_name = self.flow_name.split("\\")[-1].split(".")[0]
+        if msgbox:
+            # Show messagebox
+            import BPMN_RPA.Scripts.MessageBox as mb
+            result = mb.messagebox_show_with_yes_no_buttons(f"Continue with next step?", f"Do you want to execute the next step of flow '{instance_name}'?")
+        else:
+            # Show console
+            result = input(f"Do you want to execute the next step of flow '{instance_name}'? (y/n) ")
+            if result.lower() == "y":
+                result = True
+            else:
+                result = False
+        if result:
+            return True
+        else:
+            self.save_flow_state()
+            print(f"Flow state saved: {self.flow_name}")
+            print("Program exited.")
+            exit()
